import os

from ..rc import USER_OS, VON_DEFAULT_AUTHOR, VON_PREVIEW_PATH


def make_preview(full_path: str):
    if not os.path.exists(os.path.dirname(VON_PREVIEW_PATH)):
        os.mkdir(os.path.dirname(VON_PREVIEW_PATH))
    if USER_OS == "windows":
        full_path = full_path.replace("\\", "/")  # lmao

    with open(VON_PREVIEW_PATH, "w") as f:
        print(r"\documentclass[11pt]{scrartcl}", file=f)
        print(r"\usepackage[sexy,diagrams]{evan}", file=f)
        print(r"\author{" + VON_DEFAULT_AUTHOR + "}", file=f)

        print(r"\title{VON Preview}", file=f)
        print(r"\begin{document}", file=f)
        print(r"\input{" + full_path + "}", file=f)
        print(r"\end{document}", file=f)



import logging
import sys
import webbrowser

from .. import model, view

parser = view.Parser(prog="br", description="Browse a problem on AoPS")
parser.add_argument("key", help="The key of the problem to open")


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    entry = model.getEntryByKey(opts.key)
    if entry is None:
        logging.error(opts.key + " not found")
    else:
        url = entry.url
        if url is None:
            print("No URL is provided for this problem")
            sys.exit(1)
        print(url)
        webbrowser.open(url)



import argparse
import string
from typing import Any

from .model import PickleMappingEntry, Problem
from .puid import inferPUID
from .rc import USER_OS
from .termcolors import TERM_COLOR

if USER_OS == "windows":
    from colorama import init  # type: ignore

    init()


def APPLY_COLOR(color_name: str, s: str):
    if OPTS.color is False:
        return s
    num_leading_spaces = len(s) - len(s.lstrip())
    return (
        " " * num_leading_spaces
        + TERM_COLOR[color_name]
        + s.lstrip()
        + TERM_COLOR["RESET"]
    )


def file_escape(s: str):
    s = s.replace("/", "-")
    s = s.replace(" ", "")
    s = "".join([_ for _ in s if _ in string.ascii_letters + string.digits + "-"])
    if s == "":
        s += "emptyname"
    return s


def get_author_initials(author: str) -> str:
    author_words = author.replace(",", " ,").split(" ")
    if len(author_words) == 0:
        return "?"
    elif len(author_words) == 1:
        a = author_words[0]
        if capitals := "".join(_ for _ in a if _ in string.ascii_uppercase):
            return capitals
        return a
    else:  # len(author_words) > 1

        def passes(a: str):
            return a and (
                a[0] in string.ascii_uppercase
                or a == ","
                or all(_ in string.ascii_letters + string.digits for _ in a)
            )

        return "".join(a[0] for a in author_words if passes(a))


# Arguments hacking whee
# We have _OPTS here which will pick up any parse_args()
_view_parser = argparse.ArgumentParser(add_help=False)
_view_parser.add_argument(
    "-q",
    "--quiet",
    action="store_const",
    default=False,
    const=True,
    help="Suppress some output (only with ls now).",
)  # TODO generalize
_view_parser.add_argument(
    "--nocolor",
    action="store_const",
    dest="color",
    default=True,
    const=False,
    help="Suppress color output.",
)
_view_parser.add_argument(
    "--tabs",
    action="store_const",
    dest="tabs",
    default=False,
    const=True,
    help="Uses tabs as separator for data in list-type commands.",
)
_view_parser.add_argument(
    "--brave",
    action="store_const",
    dest="brave",
    default=False,
    const=True,
    help="Show problems marked as SECRET.",
)
_view_parser.add_argument(
    "-v",
    "--verbose",
    action="store_const",
    default=False,
    const=True,
    help="More verbose displays (e.g. include problem tags).",
)

OPTS = _view_parser.parse_args([])


class Parser(argparse.ArgumentParser):
    def __init__(self, *args: Any, **kwargs: Any):
        super(Parser, self).__init__(parents=[_view_parser], *args, **kwargs)  # type: ignore

    def process(self, *args: Any, **kwargs: Any) -> argparse.Namespace:
        global OPTS
        OPTS = self.parse_args(*args, **kwargs)
        return OPTS


def getProblemString(problem: Problem, i: int | None = None):
    s = getEntryString(problem.entry, verbose=True, i=i)
    s += "\n"
    s += APPLY_COLOR("CYAN", problem.state.strip())
    return s


def getEntryString(entry: PickleMappingEntry, verbose=False, i: int | None = None):
    # SPECIAL hide brave
    if OPTS.verbose is True:
        verbose = True
    if entry.secret and not OPTS.brave:
        return APPLY_COLOR("BOLD_YELLOW", "Problem not shown")

    if OPTS.tabs is True:
        s = "\t".join([entry.source, entry.desc, entry.sortstring])
        if verbose:
            s += "\t" + " ".join(entry.tags)
        return s
    s = ""

    # SPECIAL GLOW for index number
    if i is None:
        i = entry.i
    if i is not None:
        index_string = f"{i + 1:3}"
        if "final" in entry.tags:
            s += APPLY_COLOR("YELLOW", index_string)
        elif "waltz" in entry.tags:
            s += APPLY_COLOR("GREEN", index_string)
        elif entry.url is not None and entry.used_by_otis is False:
            s += APPLY_COLOR("BOLD_RED", index_string)
        elif entry.url is not None and entry.used_by_otis is True:
            s += APPLY_COLOR("RED", index_string)
        elif entry.used_by_otis is False:  # url missing
            s += APPLY_COLOR("BG_BLUE", index_string)
        else:  # url missing, used by OTIS
            s += APPLY_COLOR("BG_MAGENTA", index_string)
        s += " "

    # source (glows for favorite or nice)
    if verbose or len(entry.source) <= 16:
        source_string: str = entry.source
    else:
        words_in_source: list[str] = entry.source.split(" ")
        source_string = ""
        for word in words_in_source:
            if word == "Shortlist":
                source_string += "Shrt. "
            elif any(_ in string.ascii_lowercase for _ in word):
                source_string += word[:1] + "."
            else:
                source_string += word + " "
        source_string = source_string.strip()
    if "favorite" in entry.tags:
        s += APPLY_COLOR("BOLD_YELLOW", source_string)
    elif "nice" in entry.tags:
        s += APPLY_COLOR("BOLD_CYAN", source_string)
    elif "good" in entry.tags:
        s += APPLY_COLOR("BOLD_MAGENTA", source_string)
    else:
        s += APPLY_COLOR("BOLD_BLUE", source_string)
    s += " " * max(1, 17 - len(source_string))

    # hardness
    if isinstance(entry.hardness, int):
        s += APPLY_COLOR("BOLD_RED", f"{entry.hardness:2}M")
        s += " "

    # author
    if entry.author is not None:
        s += APPLY_COLOR("GREEN", "[" + get_author_initials(entry.author) + "]")
        s += " "

    # the description
    s += entry.desc if verbose else entry.desc[:40]

    # sorting hashtag
    s += " " + APPLY_COLOR("RED", "#" + entry.sortstring)

    # PUID, author credits
    if verbose:
        s += "\n" + " " * 4
        s += "PUID:" + APPLY_COLOR("CYAN", inferPUID(entry.source))
        if entry.author is not None:
            s += APPLY_COLOR("RESET", " | ")
            s += APPLY_COLOR("GREEN", entry.author)
        if entry.url is not None:
            url = entry.url
            if url.startswith("http://"):
                url = url[7:]
            if url.startswith("https://"):
                url = url[8:]
            if url.startswith("www."):
                url = url[4:]
            if len(url) > 32:
                url = url[:29] + "..."
            s += " | "
            s += APPLY_COLOR("BG_BLUE", url)

    # tags
    if verbose:
        s += "\n" + " " * 4
        s += APPLY_COLOR("MAGENTA", " ".join(entry.tags))
    return s


def formatPath(path: str):
    return "VON/" + path


def getDirString(path: str):
    return "Directory " + APPLY_COLOR("BOLD_BLUE", path)


def printProblem(*args: Any, **kwargs: Any):
    print(getProblemString(*args, **kwargs))


def printEntry(*args: Any, **kwargs: Any):
    print(getEntryString(*args, **kwargs))


def printDir(*args: Any, **kwargs: Any):
    print(getDirString(*args, **kwargs))


def out(msg: str):
    print(msg)



import logging

from .. import model, strparse, view
from ..fzf import fzf_choose

parser = view.Parser(
    prog="markup",
    description="Gives a Markdown version of problem statement for copy-pasting in email etc.",
)
parser.add_argument(
    "keys",
    nargs="*",
    help="The key(S) of the problem to open.",
)
parser.add_argument(
    "-p",
    "--preserve",
    action="store_true",
    help="Suppress macro expansion from body.",
)
parser_display_modes = parser.add_mutually_exclusive_group()

parser_display_modes.add_argument(
    "-a",
    "--aops",
    action="store_true",
    help="Returns in AoPS mode.",
)
parser_display_modes.add_argument(
    "-d",
    "--discord",
    action="store_true",
    help="Returns in Discord mode.",
)


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    keys = opts.keys
    if len(keys) == 0:
        keys = [fzf_choose()]

    for i, key in enumerate(keys):
        entry = model.getEntryByKey(key)

        if entry is None:
            logging.error(opts.key + " not found")
            return
        elif entry.secret and not opts.brave:
            logging.error("Problem can't be shown without --brave option")
            return

        problem = entry.full
        statement = problem.bodies[0]

        if opts.aops is True:
            header = "[b]"
            if problem.url is not None:
                header += f"[url={problem.url}]{entry.source}[/url]"
            else:
                header += entry.source
            if problem.author is not None:
                header += f", proposed by {problem.author}.[/b]"
            else:
                header += ".[/b]"
            view.out(header)
            view.out(strparse.toAOPS(statement))
            if i != len(keys) - 1:
                view.out("\n--------\n")
        else:
            header = "## "
            if problem.url is not None:
                header += f"[{problem.source}]({problem.url})"
            else:
                header += entry.source
            if problem.author is not None:
                header += f", proposed by {problem.author}."
            else:
                header += "."
            view.out(header)
            if opts.discord is True:
                view.out(r"```latex")
            else:
                view.out("")
            view.out(
                statement if opts.preserve is True else strparse.demacro(statement)
            )
            if opts.discord is True:
                view.out(r"```")
            elif i != len(keys) - 1:
                view.out("\n")



import datetime
import logging
import os
import re
import subprocess
import tempfile
import traceback
from argparse import Namespace
from typing import Any, Callable

import yaml
from yaml.scanner import ScannerError

from .. import model, view
from ..clipboard import get_clipboard
from ..puid import inferPUID
from ..rc import EDITOR, NSEPARATOR, SEPARATOR, TAG_HINT_TEXT
from . import preview

# https://urlregex.com/
RE_URL = re.compile(
    r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)


def user_file_input(
    initial="",
    extension=".tmp",
    pre_hook: Callable[[str], None] = None,
    delete: bool = False,
):
    """Opens in $EDITOR a file with content 'initial'
    and 'extension', and returns edited file.
    If pre_hook is not None, runs pre_hook(tf.name) before opening EDITOR.
    If delete is True, delete the file afterwards.
    """

    with tempfile.NamedTemporaryFile(suffix=extension, delete=False) as tf:
        tf.write(initial.encode())
        filename = tf.name
    if pre_hook is not None:
        pre_hook(filename)
    subprocess.run([EDITOR, filename])

    with open(filename, "r") as tf:
        edited_message = "".join(_ for _ in tf.readlines())
    if delete:
        os.unlink(filename)
    return edited_message


def alert_error_tryagain(message=""):
    """Prints an error message and waits for user to confirm."""
    logging.error(message)
    return input("** Press enter to continue: ")


PS_INSTRUCT = r"""% Input your problem and solution below.
% Three dashes on a newline indicate the breaking points.
% URL detected: {url}
% TEX root = {preview_path}
"""


def solicit_user_for_content(raw_text: str, url: str, opts: Namespace):
    del opts
    initial = (
        PS_INSTRUCT.format(
            url=url if url != "<++>" else "None",
            preview_path=preview.VON_PREVIEW_PATH,
        )
        + NSEPARATOR
        + raw_text
    )

    def pre_hook(tempfile_name: str):
        preview.make_preview(tempfile_name)

    while True:
        # TODO maybe give user instructions
        raw_ps = user_file_input(
            initial=initial, extension="von.tex", pre_hook=pre_hook
        )

        if raw_ps.count(SEPARATOR) >= 1:
            bodies = [_.strip() for _ in raw_ps.split(SEPARATOR)[1:]]
            if bodies[0] == "":
                return None
            return bodies
        elif raw_ps.strip() == "":
            return None
        else:
            alert_error_tryagain("Bad format: can't find separator. Try again.")
            initial = raw_ps


DEFAULT_PATH = model.getcwd()
YAML_DATA_FILE = """# Input your problem metadata here

source: {source}     # e.g. USAMO 2000/6. This must be unique
desc:   <++>     # e.g. Fiendish inequality
path:   {path}<++>
tags:   [{now.year}-{now.month:02d}, <++>]
url: {url}

{hint}"""


def solicit_user_for_yaml(opts: Namespace, url: str) -> None | tuple[str, Any]:
    if clipboard_text := get_clipboard():
        if RE_URL.fullmatch(clipboard_text) is not None:
            url = clipboard_text

    initial = YAML_DATA_FILE.format(
        path=model.completePath(DEFAULT_PATH),
        now=datetime.datetime.now(),
        source="<++>" if opts.source is None else opts.source,
        hint=TAG_HINT_TEXT,
        url=url,
    )
    while True:
        raw_yaml = user_file_input(initial=initial, extension="von.yaml", delete=True)
        try:
            d = yaml.safe_load(raw_yaml)
            if d is None:
                return None
            assert "path" in d, "Path is mandatory"
            assert "source" in d, "Source is mandatory"
            if d["path"][-1] != "/":
                d["path"] += "/"
            assert os.path.isdir(d["path"]), d["path"] + " directory non-existent"
            target = d["path"] + inferPUID(d["source"]) + ".tex"
            assert not os.path.isfile(target), target + " already taken"
            assert model.getEntryByKey(d["source"]) is None, (
                d["source"] + " is already an existing problem source"
            )
        except AssertionError:
            traceback.print_exc()
            alert_error_tryagain("Assertions failed, please try again.")
            initial = raw_yaml
        except ScannerError:
            traceback.print_exc()
            alert_error_tryagain("Could not parse YAML, please try again.")
            initial = raw_yaml
        else:
            del d["path"]
            # darn PyYAML used to do this fine -_-
            tags = d.pop("tags")
            output = (
                yaml.dump(d, default_flow_style=False).strip()
                + "\n"
                + "tags: ["
                + ", ".join(tags)
                + "]"
            )
            return (target, output)


def do_add_problem(raw_text: str, url: str, opts: Namespace):
    """Core procedure. Opens two instances of editors to solicit user input
    on problem and produce a problem instance."""

    # Get problem and solution
    bodies = solicit_user_for_content(raw_text, url, opts)
    if bodies is None:
        logging.warning("Aborting due to empty input...")
        return

    yaml_info = solicit_user_for_yaml(opts, url)
    if yaml_info is None:
        logging.warning("Aborting due to empty input...")
        return

    target, out_yaml = yaml_info
    out_text = NSEPARATOR.join([out_yaml] + bodies)
    p = model.addProblemByFileContents(target, out_text)
    assert p is not None
    model.augmentCache(p.entry)
    view.printEntry(p.entry)


parser = view.Parser(prog="add", description="Adds a problem to VON.")
parser.add_argument(
    "source",
    default=None,
    nargs="?",
    help="If specified, sets the source for the new problem.",
)
parser.add_argument(
    "-f",
    "--file",
    dest="filename",
    default=None,
    help="If specified, uses contents of file as body",
)
parser.add_argument(
    "-u",
    "--url",
    default=None,
    help="Specify a URL on the command line",
)


def main(self: object, argv: list[str]):
    del self
    opts = parser.process(argv)
    opts.verbose = True
    url = opts.url if opts.url is not None else "<++>"
    if opts.filename is not None:
        if not os.path.isfile(opts.filename):
            logging.error("The file " + opts.filename + " doesn't exist")
            return
        with open(opts.filename) as f:
            initial_text = "".join(f.readlines())
    else:
        if clipboard_text := get_clipboard():
            if RE_URL.fullmatch(clipboard_text) is not None and opts.url is not None:
                initial_text = "<++>"
                url = clipboard_text
            else:
                initial_text = clipboard_text
        else:
            initial_text = "<++>"
    do_add_problem(initial_text, url, opts)



import re


# Demacro
def demacro(text: str) -> str:
    # TODO this doesn't quite work, but oh well
    replacements: list[tuple[str, str]] = [
        (r"\ii ", r"\item "),
        (r"\ii[", r"\item["),
        (r"\wh", r"\widehat"),
        (r"\wt", r"\widetilde"),
        (r"\ol", r"\overline"),
        (r"\epsilon", r"\eps"),
        (r"\eps", r"\varepsilon"),
        (r"\dang", r"\measuredangle"),
        (r"\dg", r"^{\circ}"),
        (r"\inv", r"^{-1}"),
        (r"\half", r"\frac{1}{2}"),
        (r"\GL", r"\operatorname{GL}"),
        (r"\SL", r"\operatorname{SL}"),
        (r"\NN", r"{\mathbb N}"),
        (r"\ZZ", r"{\mathbb Z}"),
        (r"\CC", r"{\mathbb C}"),
        (r"\RR", r"{\mathbb R}"),
        (r"\QQ", r"{\mathbb Q}"),
        (r"\FF", r"{\mathbb F}"),
        (r"\ts", r"\textsuperscript"),
        (r"\opname", r"\operatorname"),
        (r"\defeq", r"\overset{\text{def}}{=}"),
        (r"\id", r"\operatorname{id}"),
        (r"\ord", r"\operatorname{ord}"),
        (r"\sign", r"\operatorname{sign}"),
        (r"\injto", r"\hookrightarrow"),
        (r"\vdotswithin=", r"\vdots"),
    ]
    s = text
    for short, full in replacements:
        s = s.replace(short, full)
    return s


def remove_soft_newlines(text: str) -> str:
    return re.sub(r"[a-zA-Z]\n[a-zA-Z]", lambda m: m.group(0).replace("\n", " "), text)


def toAOPS(text: str) -> str:
    DIVIDER = "\n" + r"-------------------" + "\n\n"
    text = demacro(text)
    text = text.replace(r"\qedhere", "")
    text = text.replace(r"\begin{asy}", "\n" + "[asy]" + "\n")
    text = text.replace(r"\end{asy}", "\n" + "[/asy]")
    text = text.replace(r"\begin{center}", "")
    text = text.replace(r"\end{center}", "")
    text = text.replace(r"\par ", "\n")
    text = text.replace(r"\item ", "[*]")
    text = text.replace(r"\begin{enumerate}", "[list=1]")
    text = text.replace(r"\end{enumerate}", "[/list]")
    text = text.replace(r"\begin{itemize}", "[list]")
    text = text.replace(r"\end{itemize}", "[/list]")
    for env in [
        "theorem",
        "claim",
        "lemma",
        "proposition",
        "corollary",
        "definition",
        "remark",
    ]:
        text = text.replace(
            r"\begin{" + env + "*}",
            "\n\n" + "[b][color=red]" + env.title() + ":[/color][/b] ",
        )
        text = text.replace(r"\end{" + env + "*}", "")
        text = text.replace(
            r"\begin{" + env + "}",
            "\n\n" + "[b][color=red]" + env.title() + ":[/color][/b] ",
        )
        text = text.replace(r"\end{" + env + "}", "")
    text = text.replace(r"\begin{proof}", "[i]Proof.[/i] ")
    text = text.replace(r"\end{proof}", r"$\blacksquare$" + "\n")
    text = text.replace(r"\bigskip", DIVIDER)
    text = text.replace(r"\medskip", DIVIDER)
    text = text.replace(r"\#", "#")
    text = text.replace("%\n", "\n")  # strip trailing percent signs
    # Remove Asy opacities, doesn't work on AoPS
    text = re.sub(r"opacity\(0.[0-9]+\)+([^,]+), ", "invisible, ", text)
    # Replace \emph, \textit, et al
    text = re.sub(r"\\emph{([^}]*)}", r"[i]\1[/i]", text)
    text = re.sub(r"\\textit{([^}]*)}", r"[i]\1[/i]", text)
    text = re.sub(r"\\textbf{([^}]*)}", r"[b]\1[/b]", text)
    text = re.sub(
        r"\\paragraph{([^}]*)}", DIVIDER + r"[color=blue][b]\1[/b][/color]", text
    )
    text = re.sub(r"\\subparagraph{([^}]*)}", DIVIDER + r"[b]\1[/b]", text)
    text = re.sub(r"\\url{([^}]*)}", r"[url]\1[/url]", text)
    text = re.sub(r"\\href{([^}]*)}{([^}]*)}", r"[url=\1]\2[/url]", text)
    text = text.replace(r"\arc", r"\widehat")

    # Join together newlines
    paragraphs = [
        " ".join([line.strip() for line in paragraph.splitlines()]).strip()
        for paragraph in text.split("\n\n")
    ]
    return "\n".join(paragraphs)



try:
    import pyperclip
except ModuleNotFoundError:
    PYPERCLIP_AVAILABLE = False
    pyperclip = None
else:
    PYPERCLIP_AVAILABLE = True


def get_clipboard() -> str:
    if PYPERCLIP_AVAILABLE:
        assert pyperclip is not None
        return pyperclip.paste().strip()
    else:
        return ""


def set_clipboard(s: str):
    if PYPERCLIP_AVAILABLE:
        assert pyperclip is not None
        pyperclip.copy(s)



import collections
import collections.abc
import functools
import json
import logging
import os
import pickle as pickle
import random
from typing import Any, Callable, TextIO

import yaml

from .puid import inferPUID
from .rc import (  # NOQA
    OTIS_EVIL_JSON_PATH,
    SEPARATOR,
    SORT_TAGS,
    VON_BASE_PATH,
    VON_CACHE_PATH,
    VON_INDEX_PATH,
)

OTIS_USED_SOURCES_LIST: list[str] | None
if OTIS_EVIL_JSON_PATH is not None:  # type: ignore
    with open(OTIS_EVIL_JSON_PATH) as f:
        evil_json = json.load(f)
        OTIS_HANDOUT_USED_SOURCES = evil_json.values()
else:
    OTIS_HANDOUT_USED_SOURCES = None


def shortenPath(path: str):
    return os.path.relpath(path, VON_BASE_PATH)


def completePath(path: str):
    return os.path.join(VON_BASE_PATH, path)


def vonOpen(path: str, *args: Any, **kwargs: Any) -> TextIO:
    return open(completePath(path), *args, **kwargs)


class pickleObj(collections.abc.MutableMapping):
    def _initial(self) -> Any:
        return {}

    def __init__(self, path: str, mode="rb"):
        if not os.path.isfile(path) or os.path.getsize(path) == 0:
            self.store = self._initial()
        else:
            with vonOpen(path, "rb") as f:
                self.store = pickle.load(f)  # type: ignore
        self.path = path
        self.mode = mode

    def __enter__(self):
        return self

    def __exit__(self, *_: Any):
        if self.mode == "wb":
            with vonOpen(self.path, "wb") as f:
                pickle.dump(self.store, f)  # type: ignore

    def __getitem__(self, key: Any):
        try:
            return self.store[key]
        except IndexError:
            raise IndexError(f"{key} not a valid key")

    def __setitem__(self, key: int | str, value: Any):
        if isinstance(self.store, list):
            assert isinstance(key, int)
            self.store[key] = value
        else:
            assert isinstance(key, str)
            self.store[key] = value

    def __delitem__(self, key: str):
        if isinstance(self.store, list):
            assert isinstance(key, int)
            del self.store[key]
        else:
            assert isinstance(key, str)
            del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def set(self, store: Any):
        self.store = store


class pickleDictVonIndex(pickleObj):
    store: dict[str, "PickleMappingEntry"]
    __getitem__: Callable[..., "PickleMappingEntry"]

    def _initial(self) -> dict[str, "PickleMappingEntry"]:
        return {}


class pickleListVonCache(pickleObj):
    store: list["PickleMappingEntry"]

    def __getitem__(self, idx: int) -> "PickleMappingEntry":
        return super().__getitem__(idx)

    def _initial(self) -> list["PickleMappingEntry"]:
        return []

    def set(self, store: list["PickleMappingEntry"]):
        for i in range(len(store)):
            store[i].i = i
        self.store = store


def VonIndex(mode="rb"):
    return pickleDictVonIndex(VON_INDEX_PATH, mode)


def VonCache(mode="rb"):
    return pickleListVonCache(VON_CACHE_PATH, mode)


@functools.total_ordering
class GenericItem:  # superclass to Problem, PickleMappingEntry
    desc = ""  # e.g. "Fiendish inequality"
    source = ""  # used as problem ID, e.g. "USAMO 2000/6"
    tags: list[str] = []  # tags for the problem
    path = ""  # path to problem TeX file
    i: int | None = None  # position in Cache, if any
    author: str | None = None  # default
    hardness: int | None = None  # default
    url: str | None = None

    @property
    def n(self):
        return self.i + 1 if self.i is not None else None

    @property
    def sortvalue(self):
        for i, d in enumerate(SORT_TAGS):
            if d in self.tags:
                return i
        return -1

    @property
    def sortstring(self):
        for d in SORT_TAGS:
            if d in self.tags:
                return d
        return "NONE"

    @property
    def sortkey(self):
        if isinstance(self.hardness, int):
            return (self.sortvalue, self.hardness, self.source)
        else:
            return (self.sortvalue, -1, self.source)

    @property
    def used_by_otis(self):
        if OTIS_HANDOUT_USED_SOURCES is not None:
            if "waltz" in self.tags:
                if self.source in OTIS_HANDOUT_USED_SOURCES:
                    logging.critical(f"{self.source} in both OTIS handout and exam")
                return True
            return self.source in OTIS_HANDOUT_USED_SOURCES
        else:
            return False

    def __eq__(self, other: "GenericItem") -> bool:
        return self.sortkey == other.sortkey

    def __lt__(self, other: "GenericItem") -> bool:
        return self.sortkey < other.sortkey


class Problem(GenericItem):
    bodies: list[str] = []  # statement, sol, comments, ...

    def __init__(self, path: str, **kwargs: Any):
        self.path = path
        for key in kwargs:
            setattr(self, key, kwargs[key])

    @property
    def state(self) -> str:
        return self.bodies[0]

    def __repr__(self):
        return self.source

    @property
    def entry(self) -> "PickleMappingEntry":
        """Returns an PickleMappingEntry for storage in pickle"""
        return PickleMappingEntry(
            source=self.source,
            desc=self.desc,
            author=self.author,
            url=self.url,
            hardness=self.hardness,
            tags=self.tags,
            path=self.path,
            i=self.i,
        )

    @property
    def full(self) -> "Problem":
        logging.warn("Sketchy af")
        return self


class PickleMappingEntry(GenericItem):
    def __init__(self, **kwargs: Any):
        for key in kwargs:
            if kwargs[key] is not None:
                setattr(self, key, kwargs[key])

    # search things
    def hasTag(self, tag: str):
        return tag.lower() in [_.lower() for _ in self.tags]

    def hasTerm(self, term: str):
        blob = self.source + " " + self.desc
        if self.author is not None:
            blob += " " + self.author
        return (
            term.lower() in blob.lower()
            or term in self.tags
            or term.upper() in inferPUID(self.source)
        )

    def hasAuthor(self, name: str):
        if self.author is None:
            return False
        haystacks = self.author.lower().strip().split(" ")
        return name.lower() in haystacks

    def hasSource(self, source: str):
        return source.lower() in self.source.lower()

    def __repr__(self):
        return self.source

    @property
    def secret(self):
        return "SECRET" in self.source or "secret" in self.tags

    @property
    def entry(self):
        logging.warn("sketchy af")
        return self

    @property
    def full(self) -> Problem:
        p = makeProblemFromPath(self.path)
        assert p is not None
        return p


def getcwd():
    true_dir = os.getcwd()
    if true_dir.startswith(VON_BASE_PATH) and true_dir != VON_BASE_PATH:
        return os.path.relpath(true_dir, VON_BASE_PATH)
    else:
        return ""


def getCompleteCwd():
    return completePath(getcwd())


def makeProblemFromPath(path: str) -> Problem:
    # Creates a problem instance from a source, without looking at Index
    with vonOpen(path, "r") as f:
        text = "".join(f.readlines())
    x = text.split(SEPARATOR)
    data = yaml.safe_load(x[0])
    assert data is not None, f"No data in {path}"
    data["bodies"] = [_.strip() for _ in x[1:]]
    assert data["source"], f"No source in {path}"
    assert data["desc"], f"No description in {path}"
    return Problem(path, **data)


def getAllProblems() -> list[Problem]:
    ret: list[Problem] = []
    for root, _, filenames in os.walk(VON_BASE_PATH):
        for fname in filenames:
            if not fname.endswith(".tex"):
                continue
            path = shortenPath(os.path.join(root, fname))
            p = makeProblemFromPath(path)
            ret.append(p)
    return ret


def getEntryByCacheNum(n: int) -> PickleMappingEntry:
    with VonCache() as cache:
        entry = cache[n - 1]
        return entry


def getEntryByTerm(term: str) -> PickleMappingEntry | None:
    with VonIndex() as index:
        entry = None
        if term in index:
            entry = index[term]

        if entry is None:
            puid = term.upper()
            for index_source, index_entry in index.store.items():
                if puid == inferPUID(index_source):
                    entry = index_entry
                    break

    if entry is not None:
        with VonCache() as cache:
            for cache_entry in cache:
                if cache_entry.source == entry.source:
                    return cache_entry

    return entry


def getEntryByKey(key: str):
    # TODO this shouldn't actually be in model, but blah
    if key.isdigit():
        return getEntryByCacheNum(int(key))
    else:
        return getEntryByTerm(key)


def addProblemByFileContents(path: str, text: str):
    with vonOpen(path, "w") as f:
        print(text, file=f)
    logging.info("Wrote to " + path)
    # Now update cache
    p = makeProblemFromPath(shortenPath(path))
    addProblemToIndex(p)
    return p


def viewDirectory(path: str):
    problems: list["Problem"] = []
    dirs: list[str] = []
    for item_path in os.listdir(path):
        abs_item_path = os.path.join(path, item_path)
        if os.path.isfile(abs_item_path) and abs_item_path.endswith(".tex"):
            problem = makeProblemFromPath(abs_item_path)
            assert problem is not None
            problems.append(problem)
        elif os.path.isdir(abs_item_path):
            dirs.append(item_path)
        else:
            pass  # not TeX or directory
    dirs.sort()
    entries = [p.entry for p in problems]
    entries.sort()
    if len(entries) > 0:
        setCache(entries)
    return (entries, dirs)


def runSearch(
    terms: list[str] | None = None,
    tags: list[str] | None = None,
    sources: list[str] | None = None,
    authors: list[str] | None = None,
    path="",
    refine=False,
    alph_sort=False,
    in_otis: bool | None = None,
    has_url: bool | None = None,
) -> list[PickleMappingEntry]:
    def _lambda_is_matching(entry: PickleMappingEntry):
        if OTIS_HANDOUT_USED_SOURCES is not None:
            if entry.used_by_otis and in_otis is False:
                return False
            elif not entry.used_by_otis and in_otis is True:
                return False

        if has_url is not None:
            if entry.url is None and has_url is True:
                return False
            if entry.url is not None and has_url is False:
                return False

        return all(
            (
                entry.path.startswith(path),
                (not tags or all([entry.hasTag(_) for _ in tags])),
                (not terms or all([entry.hasTerm(_) for _ in terms])),
                (not sources or all([entry.hasSource(_) for _ in sources])),
                (not authors or all([entry.hasAuthor(_) for _ in authors])),
            )
        )

    if refine is False:
        with VonIndex() as index:
            result: list[PickleMappingEntry] = [
                entry for entry in index.values() if _lambda_is_matching(entry)
            ]
    else:
        with VonCache() as cache:
            result = [entry for entry in cache.values() if _lambda_is_matching(entry)]
    if alph_sort:
        result.sort(key=lambda e: e.source)
    else:
        result.sort()
    if len(result) > 0:
        setCache(result)
    return result


def augmentCache(*entries: PickleMappingEntry):
    with VonCache("wb") as cache:
        cache.set(cache.store + list(entries))


def setCache(entries: list[PickleMappingEntry]):
    with VonCache("wb") as cache:
        cache.set(entries)


def clearCache():
    with VonCache("wb") as cache:
        cache.set([])


def readCache():
    with VonCache() as cache:
        return cache


# A certain magical Index~ <3


def addEntryToIndex(entry: PickleMappingEntry):
    with VonIndex("wb") as index:
        index[entry.source] = entry


def updateEntryByProblem(old_entry: PickleMappingEntry, new_problem: Problem):
    new_problem.i = old_entry.i
    new_entry = new_problem.entry

    with VonIndex("wb") as index:
        if old_entry.source != new_entry.source:
            del index[old_entry.source]
        index[new_entry.source] = new_entry
    with VonCache("wb") as cache:
        for i, entry in enumerate(cache):
            if entry.source == old_entry.source:
                new_entry.i = i
                cache[i] = new_entry
                break
        else:
            cache.set(cache.store + [new_entry])
    return index[new_entry.source]


def addProblemToIndex(problem: Problem):
    with VonIndex("wb") as index:
        p = problem
        index[p.source] = p.entry
        return index[p.source]


def setEntireIndex(von_index_dict: dict[str, PickleMappingEntry]):
    with VonIndex("wb") as index:
        index.set(von_index_dict)


def rebuildIndex():
    d: dict[str, PickleMappingEntry] = {}
    for p in getAllProblems():
        if p.source in d:
            fake_source = f"DUPLICATE {random.randrange(10**6, 10**7)}"
            logging.error(
                p.source + " is being repeated, replacing with " + fake_source
            )
            p.source = fake_source
        d[p.source] = p.entry
    setEntireIndex(d)



import logging
import subprocess

from .. import model, view
from ..fzf import fzf_choose
from ..rc import EDITOR
from . import preview

parser = view.Parser(prog="edit", description="Opens problem(s) by source name.")
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    dest="all",
    help="Edit all problems in cache",
)
parser.add_argument(
    "keys",
    nargs="*",
    help="The key of the problem to open (either source or cache index).",
)


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    opts.verbose = True

    if opts.all is True:
        entries = model.readCache()
    elif len(opts.keys) == 0:
        entries = [model.getEntryByKey(fzf_choose())]
    else:
        entries = [model.getEntryByKey(key) for key in opts.keys]
    for entry in entries:
        if entry is None:
            logging.error("Not found")
        else:
            full_path = model.completePath(entry.path)
            preview.make_preview(full_path)
            subprocess.call([EDITOR, full_path])  # do editing

            problem = model.makeProblemFromPath(entry.path)
            new_entry = model.updateEntryByProblem(old_entry=entry, new_problem=problem)
            # update cache after editing problem
            view.printEntry(new_entry)



from . import (  # NOQA
    add,
    asy,
    br,
    cd,
    clear,
    edit,
    index,
    link,
    markup,
    nuke,
    paths,
    po,
    search,
    show,
    solve,
    status,
)


class VonController:
    def do_ls(self, argv: list[str]):
        self.do_cd(["."] + argv)

    do_add = add.main
    do_asy = asy.main
    do_br = br.main
    do_cd = cd.main
    do_clear = clear.main
    do_cs = do_cd
    do_edit = edit.main
    do_f = show.main  # alias find
    do_link = link.main
    do_m = markup.main
    do_markup = markup.main
    do_index = index.main
    do_nuke = nuke.main
    do_paths = paths.main
    do_po = po.main
    do_s = search.main
    do_search = search.main
    do_show = show.main
    do_solve = solve.main
    do_solve = solve.main
    do_ss = status.main
    do_status = status.main



# For importing von externally
import sys

from . import model
from .puid import inferPUID

sys.modules["model"] = model

index = model.VonIndex().store  # get the underlying dict

source_to_puid_lookup = {inferPUID(source): source for source in index}


def has(source: str) -> bool:
    """Checks whether a given source exists in database"""
    return source in index


def has_solution(source: str) -> bool:
    """Checks whether a given source exists in database AND has a solution"""
    if not has(source):
        return False
    entry = index[source]
    return len(entry.full.bodies) > 1


def get_index(source: str, brave=False) -> model.PickleMappingEntry | None:
    """Returns the index entry for a given source, or None"""
    entry = index.get(source)

    if entry is None:
        puid = source.upper()
        for index_source, index_entry in index.items():
            if puid == inferPUID(index_source):
                entry = index_entry
                break

    if entry is not None and entry.secret:
        assert brave, f"{entry} requires security clearance"
    return entry


def get(source: str, brave=False) -> model.Problem:
    """Returns the full data for a given source"""
    entry = get_index(source, brave)
    assert entry is not None, f"{source} is None"
    return entry.full


def get_statement(source: str, brave=False) -> str:
    """Returns just the problem statement for a given source"""
    return get(source, brave).bodies[0]


def get_solution(source: str, brave=False) -> str:
    """Returns just the solution for a given source (asserts existence)"""
    bodies = get(source, brave).bodies
    assert len(bodies) > 1, f"{source} has no solution"
    return bodies[1]


def get_puid(source: str) -> str:
    return inferPUID(source)


def get_source(puid: str) -> str | None:
    return source_to_puid_lookup.get(puid)



import logging
import re

from .. import model, strparse, view

parser = view.Parser(
    prog="solve", description=r"Taking as input a TeX file, expands \von instances."
)

parser.add_argument("filename", help="The filename to translate.")
parser.add_argument(
    "-p",
    "--pagebreaks",
    action="store_true",
    help="Include page break after every solution (default is a bar).",
)
parser.add_argument(
    "-l",
    "--lazy",
    action="store_true",
    help="Don't include solutions to the problems.",
)
parser.add_argument(
    "-k",
    "--sourced",
    action="store_true",
    help="Always include the keyed source anyways.",
)

r = re.compile(r"\\von(\*)?(\[([^\]]+)\])?\{([A-Za-z0-9 /\-?,.!]+)\}")


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    s = ""

    with open(opts.filename) as f:
        for line in f:
            result = r.match(line)
            if result is None:
                s += line
            else:
                has_star = result.group(1) is not None
                source = result.group(2)
                key = result.group(4)
                entry = model.getEntryByKey(key)
                assert entry is not None, key
                problem = entry.full

                if has_star and not opts.sourced:
                    s += r"\begin{problem}" + "\n"
                elif source is not None:
                    if entry.url is not None and r"\href" not in source:
                        s += r"\begin{problem}["
                        s += r"\href{" + entry.url + "}{" + source[1:-1] + "}"
                        s += "]\n"
                    else:
                        s += r"\begin{problem}" + source + "\n"
                else:
                    if entry.url is not None:
                        s += r"\begin{problem}[\href{" + entry.url + "}{" + key + "}]\n"
                    else:
                        s += r"\begin{problem}[" + key + "]" + "\n"
                s += strparse.demacro(problem.bodies[0]) + "\n"
                s += r"\end{problem}" + "\n"
                if not opts.lazy:
                    if len(problem.bodies) > 1:
                        s += r"\subsection*{\ul{Solution}}" + "\n"
                        s += strparse.demacro(problem.bodies[1]) + "\n"
                        if opts.pagebreaks:
                            s += r"\newpage" + "\n"
                        else:
                            s += r"\hrulebar" + "\n"
                    else:
                        logging.error("No solution to " + key)
        view.out(s)



from pathlib import Path

from .. import model, view
from ..puid import inferPUID
from ..rc import VON_BASE_PATH

parser = view.Parser(prog="nuke", description="Fixes all the filenames to match PUID")


def main(self: object, argv: list[str]):
    parser.process(argv)
    for p in model.getAllProblems():
        puid = inferPUID(p.source)
        src = VON_BASE_PATH / Path(p.path)
        target = VON_BASE_PATH / src.parent / f"{puid}.tex"
        if not target.exists():
            src.rename(target)
    model.rebuildIndex()



import sys

from . import term
from .__about__ import __version__ as VERSION


def main():
    cmd = term.VonTerminal()
    if len(sys.argv) == 1:
        cmd.run()
    elif len(sys.argv) == 2 and sys.argv[1] in ("-h", "--help"):
        cmd.direct(["help"])
    elif len(sys.argv) == 2 and sys.argv[1] in ("-v", "--version"):
        print(VERSION)
    elif len(sys.argv) >= 2 and sys.argv[1].startswith("-"):
        print(
            f"ERROR: You need to start with a command name (rather than {sys.argv[1]})"
        )
        cmd.direct(["help"])
        sys.exit(1)
    else:
        cmd.direct(sys.argv[1:])  # to preserve command line quotes etc


if __name__ == "__main__":
    main()



import logging
import re

from .. import model, view

parser = view.Parser(
    prog="asy", description="Prints the diagram from the solution to a problem"
)
parser.add_argument("key", help="The key of the problem to open")
parser.add_argument(
    "-n",
    "--number",
    default=1,
    type=int,
    help="Takes the n-th diagram, one-indexed. Defaults to 1.",
)
parser.add_argument(
    "-b",
    "--body",
    default=1,
    type=int,
    help="Takes the diagram from the b'th body. Defaults to 1.",
)
parser.add_argument(
    "-c",
    "--comments",
    action="store_true",
    help="Decides whether or not commented lines should be displayed",
)


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    entry = model.getEntryByKey(opts.key)
    if entry is None:
        logging.error(opts.key + " not found")
    else:
        problem = entry.full
        soln = problem.bodies[opts.body]
        asys = re.findall(r"\\begin\{asy\}(.+?)\\end\{asy\}", soln, flags=re.DOTALL)
        assert opts.number >= 1
        diagram = asys[opts.number - 1]

        if not opts.comments:
            diagram = re.sub(r"/\*.*?\*/", "", diagram, flags=re.DOTALL)
        print("\n".join(line.strip() for line in diagram.strip().split("\n")))



from .. import model, view

parser = view.Parser(
    prog="paths", description="Prints the paths of all files in cache."
)


def main(self: object, argv: list[str]):
    for entry in model.readCache():
        print(entry.path)



import logging

from .. import model, view

parser = view.Parser(
    prog="search", description="Searches for problems by tags or text."
)
parser.add_argument(
    "s_terms", nargs="*", metavar="term", help="Terms you want to search for."
)
parser.add_argument(
    "-t",
    "--tag",
    nargs="+",
    metavar="tag",
    dest="s_tags",
    default=[],
    help="Tags you want to search for.",
)
parser.add_argument(
    "-k",
    "--source",
    nargs="+",
    metavar="source",
    dest="s_sources",
    default=[],
    help="Sources you want to search for.",
)
parser.add_argument(
    "-w",
    "--authors",
    nargs="+",
    metavar="authors",
    dest="s_authors",
    default=[],
    help="Authors you want to search for.",
)
parser.add_argument(
    "-r",
    "--refine",
    action="store_true",
    help="Prune through the Cache rather than the whole database.",
)
parser.add_argument(
    "-a",
    "--alphabetical",
    action="store_true",
    help="Sort the results alphabetically, not by sort tag.",
)
parser.add_argument(
    "-e", "--everything", action="store_true", help="Allow searching everything."
)
otis_group = parser.add_mutually_exclusive_group()
otis_group.add_argument(
    "-n", "--notused", action="store_true", help="Problem not used in OTIS"
)
otis_group.add_argument(
    "-o", "--occupied", action="store_true", help="Problem used in OTIS"
)
url_group = parser.add_mutually_exclusive_group()
url_group.add_argument(
    "-l", "--linked", action="store_true", help="Problem has a URL provided"
)
url_group.add_argument(
    "-u", "--unlinked", action="store_true", help="Problem has no URL provided"
)


def main(self: object, argv: list[str]):
    opts = parser.process(argv)

    query_is_empty = (
        len(opts.s_terms + opts.s_tags + opts.s_sources + opts.s_authors) == 0
    )

    if opts.everything is False and query_is_empty is True:
        logging.warning(
            "Must supply at least one search keyword or pass --everything option."
        )
        return
    if opts.everything is True and query_is_empty is False:
        logging.warning("Passing --everything with parameters makes no sense.")
        return

    if opts.notused is True:
        in_otis = False
    elif opts.occupied is True:
        in_otis = True
    else:
        in_otis = None

    if opts.unlinked is True:
        has_url = False
    elif opts.linked is True:
        has_url = True
    else:
        has_url = None

    search_path = model.getcwd()
    if search_path != "":
        logging.info(
            "Search restricted to "
            + view.APPLY_COLOR("BOLD_GREEN", view.formatPath(search_path))
        )
    result = model.runSearch(
        terms=opts.s_terms,
        tags=opts.s_tags,
        sources=opts.s_sources,
        authors=opts.s_authors,
        refine=opts.refine,
        path=search_path,
        alph_sort=opts.alphabetical,
        in_otis=in_otis,
        has_url=has_url,
    )

    for entry in result:
        view.printEntry(entry)
    if len(result) == 0:
        logging.warning("No matches found.")



from .. import model, view

parser = view.Parser(prog="ss", description="Prints the Cache.")


def main(self: object, argv: list[str]):
    parser.process(argv)
    for entry in model.readCache():
        view.printEntry(entry)



import os

from .. import model, view

parser = view.Parser(prog="cd", description="Changes the working directory.")
parser.add_argument("path", help="The path to change to")


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    os.chdir(opts.path)

    if not opts.quiet:
        entries, dirs = model.viewDirectory(model.getCompleteCwd())
        for d in dirs:
            view.printDir(d)
        for e in entries:
            view.printEntry(e)



import logging
import os

from .. import model, strparse, view
from ..fzf import fzf_choose
from ..rc import VON_DEFAULT_AUTHOR, VON_POST_OUTPUT_DIR

parser = view.Parser(prog="po", description="Prepares a LaTeX file to send to Po-Shen!")
parser.add_argument("keys", nargs="*", help="The keys of the problem to propose.")
parser.add_argument("-t", "--title", default=None, help="Title of the LaTeX document.")
parser.add_argument(
    "-s", "--subtitle", default=None, help="Subtitle of the LaTeX document."
)
parser.add_argument(
    "--author", default=VON_DEFAULT_AUTHOR, help="Author of the LaTeX document."
)
parser.add_argument("--date", default=r"\today", help="Date of the LaTeX document.")
parser.add_argument(
    "-k",
    "--sourced",
    action="store_true",
    help="Include the source.",
)
parser.add_argument(
    "--tex",
    action="store_true",
    help="Supply only the TeX source, rather than compiling to PDF.",
)
parser.add_argument(
    "-f",
    "--filename",
    default=None,
    help="Filename for the file to produce (defaults to po.tex).",
)

LATEX_PREAMBLE = r"""\usepackage{amsmath,amssymb,amsthm}
\usepackage[minimal]{yhmath}
\usepackage{derivative}

\PassOptionsToPackage{usenames,svgnames,dvipsnames}{xcolor}
\usepackage{thmtools}
\usepackage[framemethod=TikZ]{mdframed}
\usepackage{listings}

\mdfdefinestyle{mdbluebox}{%
    roundcorner = 10pt,
    linewidth=1pt,
    skipabove=12pt,
    innerbottommargin=9pt,
    skipbelow=2pt,
    linecolor=blue,
    nobreak=true,
    backgroundcolor=TealBlue!5,
}
\declaretheoremstyle[
    headfont=\sffamily\bfseries\color{MidnightBlue},
    mdframed={style=mdbluebox},
    headpunct={\\[3pt]},
    postheadspace={0pt}
]{thmbluebox}

\mdfdefinestyle{mdredbox}{%
    linewidth=0.5pt,
    skipabove=12pt,
    frametitleaboveskip=5pt,
    frametitlebelowskip=0pt,
    skipbelow=2pt,
    frametitlefont=\bfseries,
    innertopmargin=4pt,
    innerbottommargin=8pt,
    nobreak=true,
    backgroundcolor=Salmon!5,
    linecolor=RawSienna,
}
\declaretheoremstyle[
    headfont=\bfseries\color{RawSienna},
    mdframed={style=mdredbox},
    headpunct={\\[3pt]},
    postheadspace={0pt},
]{thmredbox}

\mdfdefinestyle{mdgreenbox}{%
    skipabove=8pt,
    linewidth=2pt,
    rightline=false,
    leftline=true,
    topline=false,
    bottomline=false,
    linecolor=ForestGreen,
    backgroundcolor=ForestGreen!5,
}
\declaretheoremstyle[
    headfont=\bfseries\sffamily\color{ForestGreen!70!black},
    bodyfont=\normalfont,
    spaceabove=2pt,
    spacebelow=1pt,
    mdframed={style=mdgreenbox},
    headpunct={ --- },
]{thmgreenbox}

\mdfdefinestyle{mdblackbox}{%
    skipabove=8pt,
    linewidth=3pt,
    rightline=false,
    leftline=true,
    topline=false,
    bottomline=false,
    linecolor=black,
    backgroundcolor=RedViolet!5!gray!5,
}
\declaretheoremstyle[
    headfont=\bfseries,
    bodyfont=\normalfont\small,
    spaceabove=0pt,
    spacebelow=0pt,
    mdframed={style=mdblackbox}
]{thmblackbox}

\declaretheorem[style=thmbluebox,name=Theorem]{theorem}
\declaretheorem[style=thmbluebox,name=Lemma,sibling=theorem]{lemma}
\declaretheorem[style=thmbluebox,name=Proposition,sibling=theorem]{proposition}
\declaretheorem[style=thmbluebox,name=Corollary,sibling=theorem]{corollary}
\declaretheorem[style=thmbluebox,name=Theorem,numbered=no]{theorem*}
\declaretheorem[style=thmbluebox,name=Lemma,numbered=no]{lemma*}
\declaretheorem[style=thmbluebox,name=Proposition,numbered=no]{proposition*}
\declaretheorem[style=thmbluebox,name=Corollary,numbered=no]{corollary*}

\declaretheorem[style=thmgreenbox,name=Claim,sibling=theorem]{claim}
\declaretheorem[style=thmgreenbox,name=Claim,numbered=no]{claim*}
\declaretheorem[style=thmredbox,name=Example,sibling=theorem]{example}
\declaretheorem[style=thmredbox,name=Example,numbered=no]{example*}
\declaretheorem[style=thmblackbox,name=Remark,sibling=theorem]{remark}
\declaretheorem[style=thmblackbox,name=Remark,numbered=no]{remark*}

\theoremstyle{definition}
\newtheorem{conjecture}[theorem]{Conjecture}
\newtheorem{definition}[theorem]{Definition}
\newtheorem{fact}[theorem]{Fact}
\newtheorem{ques}[theorem]{Question}
\newtheorem{exercise}[theorem]{Exercise}
\newtheorem{problem}[theorem]{Problem}

\newtheorem*{conjecture*}{Conjecture}
\newtheorem*{definition*}{Definition}
\newtheorem*{fact*}{Fact}
\newtheorem*{ques*}{Question}
\newtheorem*{exercise*}{Exercise}
\newtheorem*{problem*}{Problem}

\usepackage{mathtools}
\usepackage{hyperref}
\usepackage[shortlabels]{enumitem}
\usepackage{multirow}
\usepackage{ellipsis}

\usepackage{epic} % diagrams
\usepackage{tikz-cd} % diagrams
\usepackage{asymptote} % more diagrams
\begin{asydef}
defaultpen(fontsize(10pt));
size(8cm); // set a reasonable default
usepackage("amsmath");
usepackage("amssymb");
settings.tex="pdflatex";
settings.outformat="pdf";
// Replacement for olympiad+cse5 which is not standard
import geometry;
// recalibrate fill and filldraw for conics
void filldraw(picture pic = currentpicture, conic g, pen fillpen=defaultpen, pen drawpen=defaultpen)
    { filldraw(pic, (path) g, fillpen, drawpen); }
void fill(picture pic = currentpicture, conic g, pen p=defaultpen)
    { filldraw(pic, (path) g, p); }
// some geometry
pair foot(pair P, pair A, pair B) { return foot(triangle(A,B,P).VC); }
pair orthocenter(pair A, pair B, pair C) { return orthocentercenter(A,B,C); }
pair centroid(pair A, pair B, pair C) { return (A+B+C)/3; }
// cse5 abbreviations
path CP(pair P, pair A) { return circle(P, abs(A-P)); }
path CR(pair P, real r) { return circle(P, r); }
pair IP(path p, path q) { return intersectionpoints(p,q)[0]; }
pair OP(path p, path q) { return intersectionpoints(p,q)[1]; }
path Line(pair A, pair B, real a=0.6, real b=a) { return (a*(A-B)+A)--(b*(B-A)+B); }
// cse5 more useful functions
picture CC() {
    picture p=rotate(0)*currentpicture;
    currentpicture.erase();
    return p;
}
pair MP(Label s, pair A, pair B = plain.S, pen p = defaultpen) {
    Label L = s;
    L.s = "$"+s.s+"$";
    label(s, A, B, p);
    return A;
}
pair Drawing(Label s = "", pair A, pair B = plain.S, pen p = defaultpen) {
    dot(MP(s, A, B, p), p);
    return A;
}
path Drawing(path g, pen p = defaultpen, arrowbar ar = None) {
    draw(g, p, ar);
    return g;
}
\end{asydef}

\usepackage[headsepline]{scrlayer-scrpage}
\addtolength{\textheight}{3.14cm}
\setlength{\footskip}{0.5in}
\setlength{\headsep}{10pt}
\lehead{\normalfont\footnotesize\textbf{AUTHOR}}
\lohead{\normalfont\footnotesize\textbf{AUTHOR}}
\rehead{\normalfont\footnotesize\textbf{TITLE}}
\rohead{\normalfont\footnotesize\textbf{TITLE}}
\pagestyle{scrheadings}

\providecommand{\arc}[1]{\wideparen{#1}}
\newcommand{\hrulebar}{
\par\hspace{\fill}\rule{0.95\linewidth}{.7pt}\hspace{\fill}
\par\nointerlineskip \vspace{\baselineskip}}

\addtokomafont{paragraph}{\color{orange!35!black}\P\ }"""


def main(self: object, argv: list[str]):
    opts = parser.process(argv)

    keys = opts.keys
    if len(keys) == 0:
        keys = [fzf_choose()]

    # Better default title:
    if opts.title is not None:
        title = opts.title
    elif len(keys) == 1:
        entry = model.getEntryByKey(keys[0])
        if entry is not None:
            title = entry.source
        else:
            title = "Solution"
    else:
        title = "Solutions"

    s = r"\documentclass[11pt]{scrartcl}" + "\n"
    s += LATEX_PREAMBLE.replace("AUTHOR", opts.author).replace("TITLE", title)
    s += r"\begin{document}" + "\n"
    s += r"\title{" + title + "}" + "\n"
    if opts.subtitle is not None:
        s += r"\subtitle{" + opts.subtitle + "}" + "\n"
    s += r"\author{" + opts.author + "}" + "\n"
    s += r"\date{" + opts.date + "}" + "\n"
    s += r"\maketitle" + "\n"
    s += "\n"
    for key in keys:
        entry = model.getEntryByKey(key)
        if entry is None:
            logging.error(key + " not found")
        elif entry.secret and not opts.brave:
            logging.error(f"Problem `{entry.source}` not shown without --brave")
            return
        else:
            problem = entry.full
            s += r"\begin{problem}" if len(keys) > 1 else r"\begin{problem*}"
            if opts.sourced:
                s += "[" + entry.source + "]"
            s += "\n"
            s += strparse.demacro(problem.bodies[0]) + "\n"
            s += r"\end{problem}" if len(keys) > 1 else r"\end{problem*}"
            if entry.url:
                s += r"\noindent\emph{Link}: \url{" + entry.url + "}" + "\n"
            if len(problem.bodies) > 1:
                s += "\n" + r"\hrulebar" + "\n\n"
                s += strparse.demacro(problem.bodies[1]) + "\n"
            s += r"\pagebreak" + "\n\n"
    s += r"\end{document}"
    if opts.tex:
        view.out(s)
    else:
        if opts.filename is not None:
            fname = opts.filename
        elif len(keys) == 1:
            fname = view.file_escape(title)
        else:
            fname = "po"
        if not os.path.exists(VON_POST_OUTPUT_DIR):
            os.mkdir(VON_POST_OUTPUT_DIR)
        filepath = os.path.join(VON_POST_OUTPUT_DIR, f"{fname}.tex")
        with open(filepath, "w") as f:
            print(s, file=f)
        os.chdir(VON_POST_OUTPUT_DIR)
        os.system("latexmk -pv %s" % filepath)



import logging

from .. import model, strparse, view
from ..fzf import fzf_choose

parser = view.Parser(prog="show", description="Displays a problem by source name.")
parser.add_argument(
    "key", nargs="?", help="The key of the problem to open.", default=None
)
parser.add_argument(
    "-b",
    "--body",
    nargs="?",
    type=int,
    const=0,
    default=None,
    help="Prints only the b-th body.",
)
parser.add_argument(
    "-a",
    "--aops",
    action="store_true",
    help="Returns string in `AoPS mode'. Automatically causes -b.",
)
parser.add_argument(
    "-p",
    "--preserve",
    action="store_true",
    help="With -b, suppress macro expansion from body.",
)
parser.add_argument(
    "-t",
    "--twitch",
    action="store_true",
    help="Implies --aops --body=1 and includes an ad for Twitch Solves ISL.",
)


def main(self: object, argv: list[str]):
    opts = parser.process(argv)
    if opts.key is not None:
        entry = model.getEntryByKey(opts.key)
    else:
        entry = model.getEntryByKey(fzf_choose())
    if entry is None:
        logging.error(opts.key + " not found")
    elif entry.secret and not opts.brave:
        logging.error("Problem can't be shown without --brave option")
        return
    else:
        b = opts.body
        problem = entry.full
        if b is None and opts.twitch:
            b = 1
        elif b is None and opts.aops:
            b = 0
        elif b is None:
            view.printProblem(problem, i=entry.i)
            return
        try:
            if opts.twitch:
                view.out("Solution from [i]Twitch Solves ISL[/i]:\n")
                view.out(strparse.toAOPS(problem.bodies[b]))
            elif opts.aops:
                view.out(strparse.toAOPS(problem.bodies[b]))
            elif opts.preserve:
                view.out(problem.bodies[b])
            else:
                view.out(strparse.demacro(problem.bodies[b]))
        except IndexError:
            logging.error(f"Couldn't access {b}-th body of {problem}")



import hashlib
import re

from .rc import VON_CUSTOM_LOOKUP

lookup = {
    "Afghanistan": "AFG",
    "Albania": "ALB",
    "Algeria": "ALG",
    "Angola": "AGO",
    "Argentina": "ARG",
    "Armenia": "ARM",
    "Australia": "AUS",
    "Austria": "AUT",
    "Azerbaijan": "AZE",
    "Bahrain": "BAH",
    "Bangladesh": "BGD",
    "Belarus": "BLR",
    "Belgium": "BEL",
    "Benin": "BEN",
    "Bolivia": "BOL",
    "Bosnia and Herzegovina": "BIH",
    "Botswana": "BWA",
    "Brazil": "BRA",
    "Brunei": "BRU",
    "Bulgaria": "BGR",
    "Burkina Faso": "BFA",
    "Cambodia": "KHM",
    "Canada": "CAN",
    "Chile": "CHI",
    "China": "CHN",
    "Colombia": "COL",
    "Commonwealth of Independent States": "CIS",
    "Costa Rica": "CRI",
    "Croatia": "HRV",
    "Cuba": "CUB",
    "Cyprus": "CYP",
    "Czech": "CZE",
    "Czechoslovakia": "CZS",
    "Denmark": "DEN",
    "Dominican Republic": "DOM",
    "Ecuador": "ECU",
    "Egypt": "EGY",
    "Estonia": "EST",
    "Finland": "FIN",
    "France": "FRA",
    "Gambia": "GMB",
    "Georgia": "GEO",
    "German Democratic Republic": "GDR",
    "Germany": "GER",
    "Ghana": "GHA",
    "Greece": "HEL",
    "Guatemala": "GTM",
    "Honduras": "HND",
    "Hong Kong": "HKG",
    "Hungary": "HUN",
    "Iceland": "ISL",
    "India": "IND",
    "Indonesia": "IDN",
    "Iraq": "IRQ",
    "Iran": "IRN",
    "Ireland": "IRL",
    "Israel": "ISR",
    "Italy": "ITA",
    "Ivory Coast": "CIV",
    "Jamaica": "JAM",
    "Japan": "JPN",
    "Kazakhstan": "KAZ",
    "Kenya": "KEN",
    "North Korea": "PRK",
    "Korea": "KOR",
    "Kosovo": "KSV",
    "Kuwait": "KWT",
    "Kyrgyzstan": "KGZ",
    "Laos": "LAO",
    "Latvia": "LVA",
    "Liechtenstein": "LIE",
    "Lithuania": "LTU",
    "Luxembourg": "LUX",
    "Macau": "MAC",
    "Macedonia": "MKD",
    "Madagascar": "MDG",
    "Malaysia": "MAS",
    "Mauritania": "MRT",
    "Mexico": "MEX",
    "Moldova": "MDA",
    "Mongolia": "MNG",
    "Montenegro": "MNE",
    "Morocco": "MAR",
    "Mozambique": "MOZ",
    "Myanmar": "MMR",
    "Nepal": "NPL",
    "Netherlands": "NLD",
    "New Zealand": "NZL",
    "Nicaragua": "NIC",
    "Nigeria": "NGA",
    "North Macedonia": "MKD",
    "Norway": "NOR",
    "Oman": "OMN",
    "Pakistan": "PAK",
    "Panama": "PAN",
    "Paraguay": "PAR",
    "Peru": "PER",
    "Philippines": "PHI",
    "Poland": "POL",
    "Portugal": "POR",
    "Puerto Rico": "PRI",
    "Romania": "ROU",
    "Russia": "RUS",
    "El Salvador": "SLV",
    "Saudi Arabia": "SAU",
    "Senegal": "SEN",
    "Serbia": "SRB",
    "Serbia and Montenegro": "SCG",
    "Singapore": "SGP",
    "Slovakia": "SVK",
    "Slovenia": "SVN",
    "South Africa": "SAF",
    "Spain": "ESP",
    "Sri Lanka": "LKA",
    "Sweden": "SWE",
    "Switzerland": "SUI",
    "Syria": "SYR",
    "Taiwan": "TWN",
    "Tajikistan": "TJK",
    "Tanzania": "TZA",
    "Thailand": "THA",
    "Trinidad and Tobago": "TTO",
    "Tunisia": "TUN",
    "Turkey": "TUR",
    "Turkmenistan": "TKM",
    "Uganda": "UGA",
    "Ukraine": "UKR",
    "United Arab Emirates": "UAE",
    "United Kingdom": "UNK",
    "Uruguay": "URY",
    "Uzbekistan": "UZB",
    "Venezuela": "VEN",
    "Vietnam": "VNM",
    "Yemen": "YEM",
    "Yugoslavia": "YUG",
    "Zimbabwe": "ZWE",
}
for k, v in list(lookup.items()):
    lookup[k + " MO"] = v + "MO"
    lookup[k + " TST"] = v + "TST"
    lookup[k + " GST"] = v + "EST"  # GMO selection test
    lookup[k + " RMM TST"] = v + "RST"
    lookup[k + " EGMO TST"] = v + "EST"
    lookup[k + " JBMO TST"] = v + "JST"
    lookup[k + " TSTST"] = v + "TSTST"

lookup["AMC 10A"] = "10A"
lookup["AMC 10B"] = "10B"
lookup["AMC 12A"] = "12A"
lookup["AMC 12B"] = "12B"
lookup["ARML Local"] = "ARMLOC"
lookup["Athemath"] = "ATHE"
lookup["Balkan"] = "BALK"
lookup["Baltic Way"] = "BWAY"
lookup["Brazil Revenge"] = "BRAR"
lookup["Brazil Undergrad"] = "BRAU"
lookup["Catalunya"] = "CTLNYA"
lookup["Centroamerican"] = "CENTRO"
lookup["China TST Quiz"] = "CHNQ"
lookup["CodeForces"] = "CF"
lookup["Cono Sur"] = "CSUR"
lookup["Cyberspace Competition"] = "CYBER"
lookup["Czech Polish Slovak"] = "CPS"
lookup["December TST"] = "DECTST"
lookup["ELMO Revenge"] = "ELMOR"
lookup["ELMO SL"] = "ESL"
lookup["European Cup"] = "EURCUP"
lookup["HMMT"] = "HMMT"
lookup["Feb HMMT IntBee"] = "FEBIB"
lookup["Nov HMMT IntBee"] = "NOVIB"
lookup["CMM IntBee"] = "CMMIB"
lookup["IntBee Quals"] = "MIBQT"
lookup["IntBee"] = "MITIB"
lookup["HMNT"] = "HMNT"
lookup["Iberoamerican"] = "IBERO"
lookup["InftyDots"] = "IDOTS"
lookup["January TST"] = "JANTST"
lookup["KoMaL"] = "KML"
lookup["Kurschak"] = "KSK"
lookup["Longlist"] = "LL"
lookup["math.SE"] = "MSE"
lookup["Mathematical Reflections"] = "MR"
lookup["MathOverflow"] = "OVFW"
lookup["MathDash"] = "MD"
lookup["Math Prize"] = "MPO"
lookup["MP4G"] = "MP4G"
lookup["Napkin"] = "NAP"
lookup["NIMO Winter"] = "NIMOW"
lookup["PUMaC Finals"] = "PUF"
lookup["Putnam"] = "PTNM"
lookup["Rioplatense"] = "RPTN"
lookup["Schweitzer"] = "MSZ"
lookup["Serbia RMM TST"] = "SRBRST"
lookup["Sharygin"] = "SHRG"
lookup["Shortlist"] = "SL"
lookup["St Petersburg"] = "SPBRG"
lookup["Taiwan Quiz"] = "TWNQ"
lookup["ToT Fall"] = "TTF"
lookup["ToT Spring"] = "TTS"
lookup["Tuymaada"] = "TMD"
lookup["Twitch"] = "TWCH"
lookup["USAMO"] = "AMO"
lookup["USAMTS"] = "USMT"
lookup["USA GST"] = "USAEST"

lookup.update(VON_CUSTOM_LOOKUP)
# whooooo
REGEX = r"(?P<contest>[a-zA-Z][a-zA-Z0-9 ]+)(19|20)(?P<year>[0-9][0-9])(?P<stem>[ \/](?P<locator>[0-9A-Za-z\.\/\- ]+))?$"
re_generic = re.compile(REGEX)

sorted_lookup_keys = list(lookup.keys())
sorted_lookup_keys.sort(key=lambda x: (-len(x), x))


def getOnlyAlphanum(s: str) -> str:
    return re.sub(r"[^A-Z0-9]", "", s.upper())


def inferPUID(source: str) -> str:
    source = source.replace("Finals", "F")
    if (m := re_generic.match(source)) is not None:
        d = m.groupdict()
        contest = d["contest"].strip()
        source = (
            d["year"]
            + lookup.get(contest, getOnlyAlphanum(contest))
            + getOnlyAlphanum(d["stem"] or "")
        )
        if contest in lookup:
            return source
        thresh = 11
    elif source[0] == "H" and source[1:].isdigit():
        return source
    else:
        for k in sorted_lookup_keys:
            if source.startswith(k):
                source = lookup[k] + source[len(k) :]
                thresh = 11
                break
        else:
            thresh = 8
    source = getOnlyAlphanum(source)
    if len(source) <= thresh:
        return source
    else:
        # still too long, return some sort of hash
        return "Z" + (hashlib.sha256(source.encode("ascii")).hexdigest())[0:7].upper()



import cmd
import glob
import logging
import os
import shlex
import traceback

from . import controller, model
from .rc import USER_OS, VON_BASE_PATH
from .view import APPLY_COLOR

if USER_OS != "windows":
    import readline

    readline.set_completer_delims(" \t\n")
else:
    from pyreadline3 import Readline  # type: ignore

    readline = Readline()
    from colorama import init  # type: ignore

    init()

WELCOME_STRING = APPLY_COLOR("BOLD_YELLOW", "Welcome to VON!")
GOODBYE_STRING = APPLY_COLOR("BOLD_YELLOW", "OK, goodbye! :D")


def _complete_path(path: str):
    if os.path.isdir(path):
        return glob.glob(os.path.join(path, "*"))
    else:
        return glob.glob(path + "*")


class VonTerminal(cmd.Cmd, controller.VonController):
    def getcwd(self):
        return model.getcwd().replace(VON_BASE_PATH.rstrip("/"), "")

    @property
    def prompt(self):
        return (
            APPLY_COLOR("BOLD_CYAN", "VON/")
            + APPLY_COLOR("YELLOW", self.getcwd())
            + "\n"
            + APPLY_COLOR("BOLD_GREEN", ":)")
            + " "
        )

    def emptyline(self):
        pass

    def completedefault(self, text: str, line: str, start_idx: int, end_idx: int):
        del line, start_idx, end_idx
        return _complete_path(text)

    def run(self):
        print(WELCOME_STRING)
        os.chdir(model.getCompleteCwd())
        while 1:
            try:
                self.cmdloop()
                break
            except KeyboardInterrupt:
                print("^C")
            except SystemExit:
                pass
            except Exception:
                traceback.print_exc()
        print("\n" + GOODBYE_STRING)

    def onecmd(self, line):
        """Interpret the argument as though it had been typed in response
        to the prompt."""
        if line.strip() == "":
            return self.emptyline()
        _ = shlex.split(line)
        cmd = _[0]
        argv = _[1:]
        self.lastcmd = line
        if line == "EOF":
            self.lastcmd = ""
            return 1
        else:
            try:
                func = getattr(self, "do_" + cmd)
            except AttributeError:
                return self.default(line)
            return func(argv)

    def direct(self, cargs: list[str]):
        # cargs = sys.argv ostensibly
        if len(cargs) == 0:
            logging.error("No command given")
        cmd = cargs[0]
        if hasattr(self, "do_" + cmd):
            func = getattr(self, "do_" + cmd)
            func(cargs[1:])
        else:
            logging.error(f"Command {cmd} not recognized")

    def do_help(self, argv: list[str]):
        arg = "".join(argv)
        if arg:
            try:
                func = getattr(self, "do_" + arg)
            except AttributeError:
                logging.error(f"Command {arg} not found")
            else:
                print(APPLY_COLOR("MAGENTA", f"Getting `{arg} --help`..."))
                func(["--help"])
        else:
            print("Here is a list of available commands:")
            for name in sorted(self.get_names()):
                if name[:3] == "do_" and name != "do_help" and name != "do_EOF":
                    print("*", name[3:])
            print("Running with no arguments starts an interactive terminal.")
            print("You can get more help on a command by typing its name then --help.")
            print("To exit VON, type an EOF character (usually possible via Ctrl+D).")



import subprocess

from .model import PickleMappingEntry, VonIndex, setCache
from .puid import inferPUID


def _fzf_line(entry: PickleMappingEntry) -> str:
    puid = inferPUID(entry.source)
    return f"{puid}\t{entry.source:<13}\t{entry.desc}"


def fzf_choose() -> str:
    with VonIndex() as index:
        choices = "\n".join(_fzf_line(entry) for entry in index.values())
    chosen = subprocess.check_output(
        args=[
            "fzf",
            "-e",
            "--tabstop",
            "12",
            "-d",
            r"\t",
            "--preview",
            "python -m von show {2}",
            "--preview-window",
            "down",
        ],
        input=choices,
        text=True,
    )
    source = chosen.split("\t")[1].strip()
    setCache([index[source]])
    return source



import getpass
import os
import sys
import tempfile
from pathlib import Path
from typing import Literal

import yaml

CONFIG_DIR = (
    Path(
        os.environ.get("APPconfig")
        or os.environ.get("XDG_CONFIG_HOME")
        or os.path.join(os.environ["HOME"], ".config")
    )
    / "von"
)
CONFIG_FILE = CONFIG_DIR / "config"

EXAMPLE_CONFIG_STRING = """# vim: ft=yaml
# This is an example of a VON config file
# Please edit it to suit your needs

# REQUIRED: Choose the default author for generated TeX files.
# (You should probably put your name here)
name: "YOUR NAME HERE"

# REQUIRED: Path to folder where you want to store your VON database
# You can include "~" which gets expanded to your home directory
# Make sure this folder exists!
base_path: "~/Documents/vondb/"

# REQUIRED: A list of sort tags
# These particular tags are used for sorting and are highlighted differently.
# Specify them however you want. Should be in increasing order.
tags: ["trivial", "easy", "medium", "hard", "brutal"]

# If your terminal sucks and doesn't have color sequences (e.g. you are using Windows)
# then you should uncomment this option and change it to false
# color: true

# The following text is placed as you do editing
# in order to help you remember which tags you have.
# tag_hint_text: ""

# By default your OS is autodetected
# If this fails for some reason, you can set it to `windows`, `mac`, or `linux` below
# os: linux

# path to store temporary files (von cache, TeX previewer, posted output files)
# Defaults to the output of Python's tempfile.gettempdir()
# von_tmp_path: "/tmp"

# Name of text editor to invoke
# By default, it detects from $EDITOR automatically
# editor: vim

# If you would like to define a custom contest name (to be used in the PUID script),
# then you should define it here.
# This will also override any "standard" lookups in `puid.py`.
# abbreviations:
#   Inter Galaxy Math Olympiad: IGMO
#   Inter Universe Math Olympiad: IUMO
"""

if not CONFIG_DIR.exists():
    try:
        CONFIG_DIR.mkdir()
    except PermissionError:
        print(
            f"{CONFIG_DIR} does not exist! Could not load configuration.",
            file=sys.stderr,
        )
        sys.exit(1)

if not CONFIG_FILE.exists():
    try:
        with open(CONFIG_FILE, "w") as f:
            print(EXAMPLE_CONFIG_STRING, file=f)
    except PermissionError:
        print(
            f"{CONFIG_FILE} does not exist and could not be written!", file=sys.stderr
        )
        sys.exit(1)
    else:
        print(
            f"{CONFIG_FILE} does not exist, so we wrote a default file for you.",
            file=sys.stderr,
        )
        print(
            "You should probably open this file and change its contents.",
            file=sys.stderr,
        )
        sys.exit(1)


with open(CONFIG_FILE) as f:
    config = yaml.load(f, Loader=yaml.SafeLoader)

# Mandatory fields in config
VON_DEFAULT_AUTHOR = config["name"]
SORT_TAGS = config["tags"]
VON_BASE_PATH = str(Path(config["base_path"]).expanduser())

# Optional fields
if evil_path := config.get("evil_path", None):
    OTIS_EVIL_JSON_PATH = Path(evil_path).expanduser()
else:
    OTIS_EVIL_JSON_PATH = None

EDITOR = config.get("editor", None) or os.environ.get("EDITOR", "vim")
TAG_HINT_TEXT = config.get("tag_hint_text", "")
USE_COLOR = config.get("color", True)

# OS detector (can be overridden)
USER_OS: Literal["windows"] | Literal["mac"] | Literal["linux"]

if (user_os := config.get("os")) is not None:
    USER_OS = user_os
elif sys.platform.startswith("win32"):
    USER_OS = "windows"
elif sys.platform.startswith("darwin"):
    USER_OS = "mac"
else:
    USER_OS = "linux"  # including cygwin

VON_CUSTOM_LOOKUP = config.get("abbreviations", {})

VON_TMP_PATH = config.get("von_tmp_path", tempfile.gettempdir())
# These used to be editable but I don't think it's worth it
VON_INDEX_NAME = "index"
VON_INDEX_PATH = os.path.join(VON_BASE_PATH, VON_INDEX_NAME)
VON_CACHE_NAME = "von_cache_" + getpass.getuser()
VON_CACHE_PATH = os.path.join(VON_TMP_PATH, VON_CACHE_NAME)
VON_PREVIEW_PATH = os.path.join(
    VON_TMP_PATH, "preview_" + getpass.getuser(), "von_preview.tex"
)
VON_POST_OUTPUT_DIR = os.path.join(VON_TMP_PATH, "po_" + getpass.getuser())

SEPARATOR = "\n---\n"
NSEPARATOR = "\n" + SEPARATOR + "\n"



from .rc import USE_COLOR

TERM_COLOR: dict[str, str] = {}
TERM_COLOR["NORMAL"] = ""
TERM_COLOR["RESET"] = "\033[m"
TERM_COLOR["BOLD"] = "\033[1m"
TERM_COLOR["RED"] = "\033[31m"
TERM_COLOR["GREEN"] = "\033[32m"
TERM_COLOR["YELLOW"] = "\033[33m"
TERM_COLOR["BLUE"] = "\033[34m"
TERM_COLOR["MAGENTA"] = "\033[35m"
TERM_COLOR["CYAN"] = "\033[36m"
TERM_COLOR["BOLD_RED"] = "\033[1;31m"
TERM_COLOR["BOLD_GREEN"] = "\033[1;32m"
TERM_COLOR["BOLD_YELLOW"] = "\033[1;33m"
TERM_COLOR["BOLD_BLUE"] = "\033[1;34m"
TERM_COLOR["BOLD_MAGENTA"] = "\033[1;35m"
TERM_COLOR["BOLD_CYAN"] = "\033[1;36m"
TERM_COLOR["BG_RED"] = "\033[41m"
TERM_COLOR["BG_GREEN"] = "\033[42m"
TERM_COLOR["BG_YELLOW"] = "\033[43m"
TERM_COLOR["BG_BLUE"] = "\033[44m"
TERM_COLOR["BG_MAGENTA"] = "\033[45m"
TERM_COLOR["BG_CYAN"] = "\033[46m"

if USE_COLOR is False:  # type: ignore
    for key in list(TERM_COLOR.keys()):
        TERM_COLOR[key] = ""
